import numpy as np
from qibo import gates


def measurement_basis_rotations(hamiltonian, n_qubits, grouping=None):
    """
    Split up and sort the Hamiltonian terms to get the basis rotation gates to be applied to a quantum circuit for the
    respective (group of) terms in the Hamiltonian

    Args:
        hamiltonian (SymbolicHamiltonian): Hamiltonian (that only contains X/Y terms?)
        n_qubits: Number of qubits in the quantum circuit.
        grouping: Whether or not to group the X/Y terms together, i.e. use the same set of measurements to get the expectation
            values of a group of terms simultaneously. Default value of ``None`` will not group any terms together, which is
            the only option currently implemented.

    Returns:
        list: List of two-tuples, with each tuple given as ([`list of measurement gates`], [term1, term2, ...]), where
        term1, term2, ... are SymbolicTerms. The first tuple always corresponds to all the Z terms present, which will be two
        empty lists - ``([], [])`` - if there are no Z terms present.
    """
    result = []
    # Split up the Z and X/Y terms first
    z_only_terms = [
        term for term in hamiltonian.terms if not any(factor.name[0] in ("X", "Y") for factor in term.factors)
    ]
    xy_terms = [term for term in hamiltonian.terms if term not in z_only_terms]
    # Add the Z terms into result first
    if z_only_terms:
        result.append(([gates.M(_i) for _i in range(n_qubits)], z_only_terms))
    else:
        result.append(([], []))
    # Then add the X/Y terms in
    if xy_terms:
        if grouping is None:
            result += [
                (
                    [
                        gates.M(int(factor.target_qubit), basis=type(factor.gate))
                        for factor in term.factors
                        if factor.name[0] != "I"
                    ],
                    [
                        term,
                    ],
                )
                for term in xy_terms
            ]
        else:
            raise NotImplementedError("Not ready yet!")
    return result


def allocate_shots(grouped_terms, n_shots, method=None, max_shots_per_term=None):
    """
    Allocate shots to each group of terms in the Hamiltonian for calculating the expectation value of the Hamiltonian.

    Args:
        grouped_terms (list): Output of measurement_basis_rotations(hamiltonian, n_qubits, grouping=None
        n_shots (int): Total number of shots to be allocated
        method (str): How to allocate the shots. The available options are: ``"c"``/``"coefficients"``: ``n_shots`` is distributed
            based on the relative magnitudes of the term coefficients, ``"u"``/``"uniform"``: ``n_shots`` is distributed evenly
            amongst each term. Default value: ``"c"``.
        max_shots_per_term (int): Upper limit for the number of shots allocated to an individual group of terms. If not given,
            will be defined as a fraction (largest coefficient over the sum of all coefficients in the Hamiltonian) of ``n_shots``.

    Returns:
        list: A list containing the number of shots to be used for each group of Pauli terms respectively.
    """
    if method is None:
        method = "c"
    if max_shots_per_term is None:
        # Define based on the fraction of the term group with the largest coefficients w.r.t. sum of all coefficients.
        term_coefficients = np.array(
            [sum(abs(term.coefficient.real) for term in terms) for (_, terms) in grouped_terms]
        )
        max_shots_per_term = int(np.ceil(n_shots * (np.max(term_coefficients) / sum(term_coefficients))))
        # max_shots_per_term = min(max_shots_per_term, 250)  #  Is there an optimal value - Explore further?
    max_shots_per_term = min(n_shots, max_shots_per_term)  # Don't let max_shots_per_term > n_shots if manually defined

    n_terms = len(grouped_terms)
    shot_allocation = np.zeros(n_terms, dtype=int)

    while True:
        remaining_shots = n_shots - sum(shot_allocation)
        if not remaining_shots:
            break

        # In case n_shots is too high s.t. max_shots_per_term is too low
        # Increase max_shots_per_term, and redo the shot allocation
        if np.min(shot_allocation) == max_shots_per_term:
            max_shots_per_term = min(2 * max_shots_per_term, n_shots)
            shot_allocation = np.zeros(n_terms, dtype=int)
            continue

        if method in ("c", "coefficients"):
            # Split shots based on the relative magnitudes of the coefficients of the (group of) Pauli term(s)
            # and only for terms that haven't reached the upper limit yet
            term_coefficients = np.array(
                [
                    sum(abs(term.coefficient.real) for term in terms) if shots < max_shots_per_term else 0.0
                    for shots, (_, terms) in zip(shot_allocation, grouped_terms)
                ]
            )

            # Normalise term_coefficients, then get an initial distribution of remaining_shots
            term_coefficients /= sum(term_coefficients)
            _shot_allocation = (remaining_shots * term_coefficients).astype(int)
            # Only keep the terms with >0 shots allocated, renormalise term_coefficients, and distribute again
            term_coefficients *= _shot_allocation > 0
            if _shot_allocation.any():
                term_coefficients /= sum(term_coefficients)
                _shot_allocation = (remaining_shots * term_coefficients).astype(int)
            else:
                # For distributing the remaining few shots, i.e. remaining_shots << n_terms
                _shot_allocation = np.array(
                    allocate_shots(grouped_terms, remaining_shots, max_shots_per_term=remaining_shots, method="u")
                )

        elif method in ("u", "uniform"):
            # Uniform distribution of shots for every term. Extra shots are randomly distributed
            _shot_allocation = np.array([remaining_shots // n_terms for _ in range(n_terms)])
            if not _shot_allocation.any():
                _shot_allocation = np.zeros(n_terms)
                _shot_allocation[:remaining_shots] = 1
                np.random.shuffle(_shot_allocation)

        else:
            raise NameError("Unknown method!")

        # Add on to the current allocation, and set upper limit to the number of shots for a given term
        shot_allocation += _shot_allocation.astype(int)
        shot_allocation = np.clip(shot_allocation, 0, max_shots_per_term)

    return shot_allocation.tolist()



from qibochem.ansatz.basis_rotation import (
    br_circuit,
    givens_rotation_gate,
    givens_rotation_parameters,
    swap_matrices,
    unitary_rot_matrix,
)
from qibochem.ansatz.hardware_efficient import hea
from qibochem.ansatz.hf_reference import bk_matrix, bk_matrix_power2, hf_circuit
from qibochem.ansatz.ucc import (
    expi_pauli,
    generate_excitations,
    mp2_amplitude,
    sort_excitations,
    ucc_circuit,
)



"""
Functions for obtaining and transforming the molecular Hamiltonian
"""

import openfermion
from qibo import hamiltonians
from qibo.symbols import X, Y, Z


def fermionic_hamiltonian(oei, tei, constant):
    """
    Build molecular Hamiltonian as an InteractionOperator using the 1-/2- electron integrals

    Args:
        oei: 1-electron integrals in the MO basis
        tei: 2-electron integrals in 2ndQ notation and MO basis
        constant: Nuclear-nuclear repulsion, and inactive Fock energy if HF embedding used

    Returns:
        Molecular Hamiltonian as an InteractionOperator
    """
    oei_so, tei_so = openfermion.ops.representations.get_tensors_from_integrals(oei, tei)
    # tei_so already multiplied by 0.5, no need to include in InteractionOperator
    return openfermion.InteractionOperator(constant, oei_so, tei_so)


def qubit_hamiltonian(fermion_hamiltonian, ferm_qubit_map):
    """
    Converts the molecular Hamiltonian to a QubitOperator

    Args:
        fermion_hamiltonian: Molecular Hamiltonian as a InteractionOperator/FermionOperator
        ferm_qubit_map: Which Fermion->Qubit mapping to use

    Returns:
        qubit_operator : Molecular Hamiltonian as a QubitOperator
    """
    # Map the fermionic molecular Hamiltonian to a QubitHamiltonian
    if ferm_qubit_map == "jw":
        q_hamiltonian = openfermion.jordan_wigner(fermion_hamiltonian)
    elif ferm_qubit_map == "bk":
        q_hamiltonian = openfermion.bravyi_kitaev(fermion_hamiltonian)
    else:
        raise KeyError("Unknown fermion->qubit mapping!")
    q_hamiltonian.compress()  # Remove terms with v. small coefficients
    return q_hamiltonian


def parse_pauli_string(pauli_string, coeff):
    """
    Helper function: Converts a single Pauli string to a Qibo Symbol

    Args:
        pauli_string (tuple of tuples): Indicate what gates to apply onto which qubit
            e.g. ((0, 'Z'), (2, 'Z'))
        coeff (float): Coefficient of the Pauli string

    Returns:
        qibo.symbols.Symbol for a single Pauli string, e.g. -0.04*X0*X1*Y2*Y3
    """
    # Dictionary for converting
    xyz_to_symbol = {"X": X, "Y": Y, "Z": Z}
    # Check that pauli_string is non-empty
    if pauli_string:
        # pauli_string format: ((0, 'Y'), (1, 'Y'), (3, 'X'))
        qibo_pauli_string = 1.0
        for p_letter in pauli_string:
            qibo_pauli_string *= xyz_to_symbol[p_letter[1]](p_letter[0])
        # Include coefficient after all symbols
        qibo_pauli_string = coeff * qibo_pauli_string
    else:
        # Empty word, i.e. constant term in Hamiltonian
        qibo_pauli_string = coeff
    return qibo_pauli_string


def symbolic_hamiltonian(q_hamiltonian):
    """
    Converts a OpenFermion QubitOperator to a Qibo SymbolicHamiltonian

    Args:
        q_hamiltonian: QubitOperator

    Returns:
        qibo.hamiltonians.SymbolicHamiltonian
    """
    # Sums over each individual Pauli string in the QubitOperator
    symbolic_ham = sum(
        parse_pauli_string(pauli_string, coeff)
        # Iterate over all operators
        for operator in q_hamiltonian.get_operators()
        # .terms gives one operator as a dictionary with one entry
        for pauli_string, coeff in operator.terms.items()
    )

    # Map the QubitHamiltonian to a Qibo SymbolicHamiltonian and return it
    return hamiltonians.SymbolicHamiltonian(symbolic_ham)



from qibo import gates


def hea(n_layers, n_qubits, parameter_gates=["RY", "RZ"], coupling_gates="CZ"):
    """
    Builds the generalized hardware-efficient ansatz, in which the rotation and entangling gates used can be
    chosen by the user

    Args:
        n_layers: Number of layers of rotation and entangling gates
        n_qubits: Number of qubits in the quantum circuit
        parameter_gates: List of single-qubit rotation gates to be used in the ansatz. The gates should be given as
            strings representing valid ``Qibo`` one-qubit gates. Default: ``["RY", "RZ"]``
        coupling_gates: String representing the two-qubit entangling gate to be used in the ansatz; should be a
            valid two-qubit ``Qibo`` gate. Default: ``"CZ"``

    Returns:
        ``list``: Gates corresponding to the hardware-efficient ansatz
    """
    gate_list = []

    for ilayer in range(n_layers):
        for rgate in parameter_gates:
            for iqubit in range(n_qubits):
                gate_list.append(getattr(gates, rgate)(iqubit, theta=0))

        # entanglement
        for iqubit in range(n_qubits - 1):
            gate_list.append(getattr(gates, coupling_gates)(iqubit, iqubit + 1))
        gate_list.append(getattr(gates, coupling_gates)(n_qubits - 1, 0))

    return gate_list



"""
Circuit representing an unitary rotation of the molecular (spin-)orbital basis set
"""

import numpy as np
from openfermion.linalg.givens_rotations import givens_decomposition
from qibo import gates, models
from scipy.linalg import expm

# Helper functions


def unitary_rot_matrix(parameters, occ_orbitals, virt_orbitals):
    r"""
    Returns the unitary rotation matrix U = exp(\kappa) mixing the occupied and virtual orbitals,
         where \kappa is an anti-Hermitian matrix

    Args:
        parameters: List/array of rotation parameters. dim = len(occ_orbitals)*len(virt_orbitals)
        occ_orbitals: Iterable of occupied orbitals
        virt_orbitals: Iterable of virtual orbitals
    """
    n_orbitals = len(occ_orbitals) + len(virt_orbitals)

    kappa = np.zeros((n_orbitals, n_orbitals))
    # Get all possible (occ, virt) pairs
    orbital_pairs = [(_o, _v) for _o in occ_orbitals for _v in virt_orbitals]
    # occ/occ and virt/virt orbital mixing doesn't change the energy, so can ignore
    for _i, (_occ, _virt) in enumerate(orbital_pairs):
        kappa[_occ, _virt] = parameters[_i]
        kappa[_virt, _occ] = -parameters[_i]
    return expm(kappa)


def swap_matrices(permutations, n_qubits):
    r"""
    Build matrices that permute (swap) the columns of a given matrix

    Args:
        permutations [List(tuple, ), ]: List of lists of 2-tuples permuting two consecutive numbers
            e.g. [[(0, 1), (2, 3)], [(1, 2)], ...]
        n_qubits: Dimension of matrix (\exp(\kappa) matrix) to be operated on i.e. # of qubits

    Returns:
        List of matrices that carry out \exp(swap) for each pair in permutations
    """
    exp_swaps = []
    _temp = np.array([[-1.0, 1.0], [1.0, -1.0]])
    for pairs in permutations:
        gen = np.zeros((n_qubits, n_qubits))
        for pair in pairs:
            # Add zeros to pad out the (2, 2) matrix to (n_qubits, n_qubits) with 0.
            gen += np.pad(_temp, ((pair[0], n_qubits - pair[1] - 1),), "constant")
        # Take matrix exponential, remove all entries close to zero, and convert to real matrix
        exp_mat = expm(-0.5j * np.pi * gen)
        exp_mat[np.abs(exp_mat) < 1e-10] = 0.0
        exp_mat = np.real_if_close(exp_mat)
        # Add exp_mat to exp_swaps once cleaned up
        exp_swaps.append(exp_mat)
    return exp_swaps


def givens_rotation_parameters(n_qubits, exp_k, n_occ):
    r"""
    Get the parameters of the Givens rotation matrix obtained after decomposing \exp(\kappa)

    Args:
        n_qubits: Number of qubits (i.e. spin-orbitals)
        exp_k: Unitary rotation matrix
        n_occ: Number of occupied orbitals
    """
    # List of 2-tuples to link all qubits via swapping
    swap_pairs = [
        [(_i, _i + 1) for _i in range(_d % 2, n_qubits - 1, 2)]
        # Only 2 possible ways of swapping: [(0, 1), ... ] or [(1, 2), ...]
        for _d in range(2)
    ]
    # Get their corresponding matrices
    swap_unitaries = swap_matrices(swap_pairs, n_qubits)

    # Create a copy of exp_k for rotating
    exp_k_shift = np.copy(exp_k)
    # Apply the column-swapping transformations
    for u_rot in swap_unitaries:
        exp_k_shift = u_rot @ exp_k_shift
    # Only want the n_occ by n_qubits (n_orb) sub-matrix
    qmatrix = exp_k_shift.T[:n_occ, :]

    # Apply Givens decomposition of the submatrix to get a list of individual Givens rotations
    # (orb1, orb2, theta, phi), where orb1 and orb2 are rotated by theta(real) and phi(imag)
    qdecomp, _, _ = givens_decomposition(qmatrix)
    # Lastly, reverse the list of Givens rotations (because U = G_k ... G_2 G_1)
    return list(reversed(qdecomp))


def givens_rotation_gate(n_qubits, orb1, orb2, theta):
    """
    Circuit corresponding to a Givens rotation between two qubits (spin-orbitals)

    Args:
        n_qubits: Number of qubits in the quantum circuit
        orb1, orb2: Orbitals used for the Givens rotation
        theta: Rotation parameter

    Returns:
        circuit: Qibo Circuit object representing a Givens rotation between orb1 and orb2
    """
    # Define 2x2 sqrt(iSwap) matrix
    iswap_mat = np.array([[1.0, 1.0j], [1.0j, 1.0]]) / np.sqrt(2.0)
    # Build and add the gates to circuit
    circuit = models.Circuit(n_qubits)
    circuit.add(gates.GeneralizedfSim(orb1, orb2, iswap_mat, 0.0, trainable=False))
    circuit.add(gates.RZ(orb1, -theta))
    circuit.add(gates.RZ(orb2, theta + np.pi))
    circuit.add(gates.GeneralizedfSim(orb1, orb2, iswap_mat, 0.0, trainable=False))
    circuit.add(gates.RZ(orb1, np.pi, trainable=False))
    return circuit


def br_circuit(n_qubits, parameters, n_occ):
    """
    Google's basis rotation circuit, applied between the occupied/virtual orbitals. Forms the exp(kappa) matrix, decomposes
    it into Givens rotations, and sets the circuit parameters based on the Givens rotation decomposition. Note: Supposed
    to be used with the JW fermion-to-qubit mapping

    Args:
        n_qubits: Number of qubits in the quantum circuit
        parameters: Rotation parameters for exp(kappa); Must have (n_occ * n_virt) parameters
        n_occ: Number of occupied orbitals

    Returns:
        Qibo ``Circuit``: Circuit corresponding to the basis rotation ansatz between the occupied and virtual orbitals
    """
    assert len(parameters) == (n_occ * (n_qubits - n_occ)), "Need len(parameters) == (n_occ * n_virt)"
    # Unitary rotation matrix \exp(\kappa)
    exp_k = unitary_rot_matrix(parameters, range(n_occ), range(n_occ, n_qubits))
    # List of Givens rotation parameters
    g_rotation_parameters = givens_rotation_parameters(n_qubits, exp_k, n_occ)

    # Start building the circuit
    circuit = models.Circuit(n_qubits)
    # Add the Givens rotation circuits
    for g_rot_parameter in g_rotation_parameters:
        for orb1, orb2, theta, _phi in g_rot_parameter:
            assert np.allclose(_phi, 0.0), "Unitary rotation is not real!"
            circuit += givens_rotation_gate(n_qubits, orb1, orb2, theta)
    return circuit
